using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;

namespace BindingList_Test
{
    /// <summary>
    /// A collection of <see cref="Person" /> objects.
    /// </summary>
    public class PersonCollection : System.ComponentModel.BindingList<Person>
    {
#region Types

        /// <summary>
        /// Compares two <see cref="Person" /> objects based on a specific property.
        /// </summary>
        private class PersonComparer : System.Collections.Generic.IComparer<Person>
        {
            /// <summary>
            /// The property on which to compare.
            /// </summary>
            private PropertyDescriptor prop;
            /// <summary>
            /// The direction in which values should be sorted.
            /// </summary>
            private ListSortDirection direction;

            /// <summary>
            /// Compares two <see cref="Person" /> objects.
            /// </summary>
            /// <param name="x">
            /// The first <b>Person</b> object.
            /// </param>
            /// <param name="y">
            /// The second <b>Person</b> object.
            /// </param>
            /// <returns>
            /// An <b>Integer</b> value less than zero if <i>x</i> is relatively less than <i>y</i>;
            /// An <b>Integer</b> value greater than zero if <i>x</i> is relatively greater than <i>y</i>;
            /// Zero if <i>x</i> is relatively equivalent to <i>y</i>;
            /// </returns>
            public PersonComparer(PropertyDescriptor prop, ListSortDirection direction)
            {
                this.prop = prop;
                this.direction = direction;
            }

            /// <summary>
            /// Creates a new instance of the <see cref="PersonComparer" /> class.
            /// </summary>
            /// <param name="prop">
            /// Represents the property of the <see cref="Person" /> objects whose values should be compared
            /// </param>
            /// <param name="direction">
            /// Indicates the direction in which the objects are being sorted.
            /// </param>
            /// <remarks></remarks>
            public int Compare(Person x, Person y)
            {
                // Get a value that indicates the relative positions of the values of the specified property.
                int result = ((IComparable)this.prop.GetValue(x)).CompareTo(this.prop.GetValue(y));

                // If the sort order is descending...
                if (this.direction == ListSortDirection.Descending)
                {
                    // ...reverse the relative positions.
                    result = -result;
                }

                return result;
            }
        }

#endregion // Types

#region Variables

        /// <summary>
        /// The property the items are sorted by.
        /// </summary>
        private PropertyDescriptor _sortProperty;
        /// <summary>
        /// The direction the items are sorted in.
        /// </summary>
        /// <remarks>
        /// Ignored if the sort property is not set.
        /// </remarks>
        private ListSortDirection _sortDirection;

#endregion // Variables

#region Properties

        /// <summary>
        /// Gets or sets a clause that indicates how the collection is sorted.
        /// </summary>
        /// <value>
        /// A <b>String</b> containing a property name and a direction if the collection is sorted; otherwise, an empty string.
        /// </value>
        public string Sort
        {
            get
            {
                return this.GetSortClause();
            }
            set
            {
                PropertyDescriptor prop;
                ListSortDirection direction;

                // Parse the string, which will throw an exception on invalid data.
                this.ParseSortClause(value, out prop, out direction);

                if (prop == null)
                {
                    // The collection is unsorted.
                    this.RemoveSortCore();
                }
                else
                {
                    // Sort the collection as requested.
                    this.ApplySortCore(prop, direction);
                }
            }
        }


        /// <summary>
        /// Gets the direction in which the collection is sorted.
        /// </summary>
        /// <value>
        /// <see cref="ListSortDirection.Ascending">Ascending</see> if the collection is sorted in ascending order or isn't sorted; otherwise, <see cref="ListSortDirection.Descending">Descending</see>.
        /// </value>
        protected override ListSortDirection SortDirectionCore
        {
            get
            {
                return this._sortDirection;
            }
        }

        /// <summary>
        /// Gets the property by which the collection is sorted.
        /// </summary>
        /// <value>
        /// A <see cref="PropertyDescriptor" /> if the collection is sorted; otherwise, a null reference.
        /// </value>
        protected override PropertyDescriptor SortPropertyCore
        {
            get
            {
                return this._sortProperty;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the collection supports sorting.
        /// </summary>
        /// <value>
        /// Always returns <b>True</b>.
        /// </value>
        protected override bool SupportsSortingCore
        {
            get
            {
                return true;
            }
        }

#endregion // Properties

#region Methods

        /// <summary>
        /// Sorts the collection.
        /// </summary>
        /// <param name="prop">
        /// The property by which to sort the items.
        /// </param>
        /// <param name="direction">
        /// The direction in which to sort the items.
        /// </param>
        protected override void ApplySortCore(PropertyDescriptor prop, ListSortDirection direction)
        {
            this._sortProperty = prop;
            this._sortDirection = direction;

            // Create a array to temporarily hold the items.
            Person[] items = new Person[this.Items.Count];

            this.Items.CopyTo(items, 0);

            // Sort the array.
            Array.Sort(items,
                       new PersonComparer(prop,
                                          direction));

            // Replace the existing items with the sorted items.
            for (int index = 0; index < items.Length; index++)
            {
                this.Items[index] = items[index];
            }

            // Notify the UI of the change.
            this.OnListChanged(new ListChangedEventArgs(ListChangedType.Reset, 0));
        }

        /// <summary>
        /// Removes all elements from the collection.
        /// </summary>
        protected override void ClearItems()
        {
            // Detach event handlers from all items.
            foreach (Person item in this.Items)
            {
                item.PropertyChanged -= new PropertyChangedEventHandler(Person_PropertyChanged);
            }

            base.ClearItems();
        }

        /// <summary>
        /// Inserts the specified item in the list at the specified index.
        /// </summary>
        protected override void InsertItem(int index, Person item)
        {
            base.InsertItem(index, item);

            // Attach event handlers to the item being added.
            item.PropertyChanged += new PropertyChangedEventHandler(Person_PropertyChanged);

            if (this._sortProperty != null)
            {
                // Ensure that the collection maintains the correct sort order.
                this.ApplySortCore(this._sortProperty, this._sortDirection);
            }
        }

        /// <summary>
        /// Removes the item at the specified index.
        /// </summary>
        protected override void RemoveItem(int index)
        {
            // Detach event handlers from the item being removed.
            this.Items[index].PropertyChanged -= new PropertyChangedEventHandler(Person_PropertyChanged);

            base.RemoveItem(index);
        }

        /// <summary>
        /// Unsorts the collection.
        /// </summary>
        /// <remarks>
        /// The order of the existing items is unaffected but new items will not be inserted according to a sort order.
        /// </remarks>
        protected override void RemoveSortCore()
        {
            this._sortProperty = null;
            this._sortDirection = ListSortDirection.Ascending;
        }

        /// <summary>
        /// Replaces the item at the specified index with the specified item.
        /// </summary>
        protected override void SetItem(int index, Person item)
        {
            // Detach event handlers from the item being removed.
            this.Items[index].PropertyChanged -= new PropertyChangedEventHandler(Person_PropertyChanged);

            base.SetItem(index, item);

            // Attach event handlers to the item being added.
            item.PropertyChanged += new PropertyChangedEventHandler(Person_PropertyChanged);

            if (this._sortProperty != null)
            {
                // Ensure that the collection maintains the correct sort order.
                this.ApplySortCore(this._sortProperty, this._sortDirection);
            }
        }


        /// <summary>
        /// Gets an abbreviation representing the sort order.
        /// </summary>
        /// <returns>
        /// "ASC" if the collection is sorted in ascending order, otherwise, "DESC".
        /// </returns>
        private string GetDirectionAbbreviation()
        {
            string abbreviation = null;

            switch (this._sortDirection)
            {
                case ListSortDirection.Ascending:
                    abbreviation = "ASC";
                    break;
                case ListSortDirection.Descending:
                    abbreviation = "DESC";
                    break;
            }

            return abbreviation;
        }

        /// <summary>
        /// Gets a clause indicating how the collection is sorted.
        /// </summary>
        /// <returns>
        /// A <b>String</b> containing a property name and a direction if the collection is sorted; otherwise, an empty string.
        /// </returns>
        private string GetSortClause()
        {
            string clause;

            if (this._sortProperty == null)
            {
                clause = string.Empty;
            }
            else
            {
                clause = string.Format("{0} {1}",
                                       this._sortProperty.Name,
                                       this.GetDirectionAbbreviation());
            }

            return clause;
        }

        /// <summary>
        /// Gets the property name and direction from a sort clause.
        /// </summary>
        /// <param name="clause">
        /// The sort clause to parse.
        /// </param>
        /// <param name="prop">
        /// The property by which the collection will be sorted.
        /// </param>
        /// <param name="direction">
        /// The direction in which the collection will be sorted.
        /// </param>
        private void ParseSortClause(string clause,
                                     out PropertyDescriptor prop,
                                     out ListSortDirection direction)
        {
            // Assume no sorting.
            prop = null;

            // Use the default direction.
            direction = ListSortDirection.Ascending;

            // Split the clause on spaces and discard any empty results.
            string[] parts = clause.Split(new Char[] { ' ' },
                                          StringSplitOptions.RemoveEmptyEntries);

            if (parts.Length > 2)
            {
                throw new ArgumentException("Invalid sort clause");
            }

            if (parts.Length > 0)
            {
                // The first part is the property name so create a PropertyDescriptor.
                prop = TypeDescriptor.GetProperties(typeof(Person))[parts[0]];

                if (prop == null)
                {
                    throw new ArgumentException("Invalid property name");
                }

                if (parts.Length > 1)
                {
                    // Get the direction.
                    if (string.Compare(parts[1], "ASC", true) == 0)
                    {
                        direction = ListSortDirection.Ascending;
                    }
                    else if (string.Compare(parts[1], "DESC", true) == 0)
                    {
                        direction = ListSortDirection.Descending;
                    }
                    else
                    {
                        throw new ArgumentException("Invalid sort direction");
                    }
                }
            }
        }

#endregion // Methods

#region Event Handlers

        /// <summary>
        /// Handles the PropertyChanged event for all items in the collection.
        /// </summary>
        /// <param name="sender">
        /// The item whose property value has changed.
        /// </param>
        /// <param name="e">
        /// Contains the name of the property whose value has changed.
        /// </param>
        private void Person_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (this._sortProperty.Name == e.PropertyName)
            {
                // A value has changed in the property by which the items are sorted so resort the collection.
                this.ApplySortCore(this._sortProperty, this._sortDirection);
            }
        }

#endregion // Event Handlers
    }
}
